"""

The Remote driver uses the following packet stricture.

1 byte for packet type. "D" for data, "M" for meta.
4 byte little endian integer for the size of the payload.
Arbitrary payload.


"""

from __future__ import annotations

import asyncio
import json
import platform
import signal
from codecs import getincrementaldecoder
from threading import Event

from textual import events, log, messages
from textual._xterm_parser import XTermParser
from textual.app import App
from textual.driver import Driver as _Driver
from textual.drivers._byte_stream import ByteStream
from textual.geometry import Size

WINDOWS = platform.system() == "Windows"


class _ExitInput(Exception):
    """Internal exception to force exit of input loop."""


class Driver(_Driver):
    """A headless driver that may be run remotely."""

    def __init__(
        self,
        app: App,
        *,
        debug: bool = False,
        mouse: bool = True,
        size: tuple[int, int] | None = None,
    ):
        super().__init__(app, debug=debug, mouse=mouse, size=size)
        self.exit_event = Event()
        self._process_input_task = asyncio.create_task(self.process_input())
        self._stdout_queue = asyncio.Queue()
        self._stdin_queue = asyncio.Queue()

    def write(self, data: str) -> None:
        """Write data to the output device.

        Args:
            data: Raw data.
        """

        data_bytes = data.encode("utf-8")
        self._stdout_queue.put_nowait(b"D%s%s" % (len(data_bytes).to_bytes(4, "big"), data_bytes))

    def write_meta(self, data: dict[str, object]) -> None:
        """Write meta to the controlling process (i.e. textual-web)

        Args:
            data: Meta dict.
        """
        meta_bytes = json.dumps(data).encode("utf-8", errors="ignore")
        self._stdout_queue.put_nowait(b"M%s%s" % (len(meta_bytes).to_bytes(4, "big"), meta_bytes))

    def flush(self) -> None:
        pass

    def _enable_mouse_support(self) -> None:
        """Enable reporting of mouse events."""
        write = self.write
        write("\x1b[?1000h")  # SET_VT200_MOUSE
        write("\x1b[?1003h")  # SET_ANY_EVENT_MOUSE
        write("\x1b[?1015h")  # SET_VT200_HIGHLIGHT_MOUSE
        write("\x1b[?1006h")  # SET_SGR_EXT_MODE_MOUSE

    def _enable_bracketed_paste(self) -> None:
        """Enable bracketed paste mode."""
        self.write("\x1b[?2004h")

    def _disable_bracketed_paste(self) -> None:
        """Disable bracketed paste mode."""
        self.write("\x1b[?2004l")

    def _disable_mouse_support(self) -> None:
        """Disable reporting of mouse events."""
        write = self.write
        write("\x1b[?1000l")  #
        write("\x1b[?1003l")  #
        write("\x1b[?1015l")
        write("\x1b[?1006l")

    def _request_terminal_sync_mode_support(self) -> None:
        """Writes an escape sequence to query the terminal support for the sync protocol."""
        self.write("\033[?2026$p")

    def start_application_mode(self) -> None:
        """Start application mode."""

        loop = asyncio.get_running_loop()

        def do_exit() -> None:
            """Callback to force exit."""
            asyncio.run_coroutine_threadsafe(
                self._app._post_message(messages.ExitApp()), loop=loop
            )

        if not WINDOWS:
            for _signal in (signal.SIGINT, signal.SIGTERM):
                loop.add_signal_handler(_signal, do_exit)

        self._stdout_queue.put_nowait(b"__GANGLION__\n")

        self.write("\x1b[?1049h")  # Alt screen
        self._enable_mouse_support()

        self.write("\x1b[?25l")  # Hide cursor
        self.write("\033[?1003h\n")

        size = Size(80, 24) if self._size is None else Size(*self._size)
        event = events.Resize(size, size)
        asyncio.run_coroutine_threadsafe(
            self._app._post_message(event),
            loop=loop,
        )

        self._request_terminal_sync_mode_support()
        self._enable_bracketed_paste()
        self.flush()
        self._app.call_later(self._app.post_message, events.AppBlur())

    def disable_input(self) -> None:
        """Disable further input."""

    def stop_application_mode(self) -> None:
        """Stop application mode, restore state."""
        self.exit_event.set()
        self.write_meta({"type": "exit"})

    async def process_input(self) -> None:
        """Wait for input and dispatch events."""
        parser = XTermParser(debug=self._debug)
        utf8_decoder = getincrementaldecoder("utf-8")().decode
        decode = utf8_decoder
        # The server sends us a stream of bytes, which contains the equivalent of stdin, plus
        # in band data packets.
        byte_stream = ByteStream()
        try:
            while True:
                data = await self._stdin_queue.get()
                for packet_type, payload in byte_stream.feed(data):
                    if packet_type == "D":
                        # Treat as stdin
                        for event in parser.feed(decode(payload)):
                            self.process_message(event)
                    else:
                        # Process meta information separately
                        self._on_meta(packet_type, payload)
                for event in parser.tick():
                    self.process_message(event)
        except _ExitInput:
            pass
        except Exception:
            from traceback import format_exc

            log(format_exc())
        finally:
            self._process_input_task.cancel()

    def _on_meta(self, packet_type: str, payload: bytes) -> None:
        """Private method to dispatch meta.

        Args:
            packet_type: Packet type (currently always "M")
            payload: Meta payload (JSON encoded as bytes).
        """
        payload_map = json.loads(payload)
        _type = payload_map.get("type", {})
        if isinstance(_type, str):
            self.on_meta(_type, payload_map)
        else:
            log.error(
                f"Protocol error: type field value is not a string. Value is {_type!r}"
            )

    def on_meta(self, packet_type: str, payload: dict) -> None:
        """Process meta information.

        Args:
            packet_type: The type of the packet.
            payload: meta dict.
        """
        if packet_type == "resize":
            self._size = (payload["width"], payload["height"])
            requested_size = Size(*self._size)
            self._app.post_message(events.Resize(requested_size, requested_size))
        elif packet_type == "focus":
            self._app.post_message(events.AppFocus())
        elif packet_type == "blur":
            self._app.post_message(events.AppBlur())
        elif packet_type == "quit":
            self._app.post_message(messages.ExitApp())
        elif packet_type == "exit":
            raise _ExitInput()



from __future__ import annotations

import asyncio
import json
from typing import Type

from textual import events
from textual.app import App
from textual.geometry import Size
from ypywidgets import Reactive, Widget as _Widget
from ypywidgets.comm import CommWidget

from ._driver import Driver


class WidgetModel(_Widget):
    _data_from_app = Reactive[str]("")
    _data_to_app = Reactive[str]("")
    _cols = Reactive[int](0)
    _rows = Reactive[int](0)
    _ready = Reactive[bool](False)

    def __init__(
        self,
        ydoc=None,
    ) -> None:
        super().__init__(ydoc)


class Widget(CommWidget, WidgetModel):

    def __init__(self, app_class: Type[App]) -> None:
        CommWidget.__init__(self)
        WidgetModel.__init__(self)
        self._app = app_class(driver_class=Driver)
        self._cols = 0
        self._rows = 0
        self._data_from_app_queue = asyncio.Queue()
        self._data_to_app_queue = asyncio.Queue()
        self._tasks = [
            asyncio.create_task(self._run()),
            asyncio.create_task(self._send_data()),
            asyncio.create_task(self._recv_data()),
        ]

        @WidgetModel._ready.watch
        def _watch__ready(obj, old: bool, value: bool):
            if obj != self:
                return

            if value:
                size = Size(int(obj._cols), int(obj._rows))
                loop = asyncio.get_running_loop()
                event = events.Resize(size, size)
                asyncio.run_coroutine_threadsafe(
                    self._app._post_message(event),
                    loop=loop,
                )

        @WidgetModel._data_to_app.watch
        def _watch__data_to_app(obj, old:str, data: str):
            if obj != self:
                return

            if data:
                obj._data_to_app_queue.put_nowait(data)

    async def _send_data(self):
        while True:
            data = await self._data_from_app_queue.get()
            self._data_from_app = ""
            self._data_from_app = data.decode()

    async def _recv_data(self):
        while True:
            data = await self._data_to_app_queue.get()
            if not data:
                continue
            data = data.encode()
            packet_type = b"D"
            packet = b"%s%s%s" % (packet_type, len(data).to_bytes(4, "big"), data)
            self._app._driver._stdin_queue.put_nowait(packet)

    async def _run(self):
        self._tasks.append(asyncio.create_task(self._app.run_async()))
        while True:
            if self._app._driver is not None:
                break
            await asyncio.sleep(0)

        for _ in range(10):
            line = []
            while True:
                data = await self._app._driver._stdout_queue.get()
                line.append(data)
                if data[-1] == 10:  # "\n"
                    break
            line = b"".join(line)
            if not line:
                break
            if line == b"__GANGLION__\n":
                ready = True
                break
        if ready:
            META = 77  # b"M"
            DATA = 68  # b"D"
            while True:
                data = await self._app._driver._stdout_queue.get()
                type_bytes = data[0]
                size_bytes = data[1:5]
                size = int.from_bytes(size_bytes, "big")
                data = data[5:5 + size]
                if type_bytes == DATA:
                    self._data_from_app_queue.put_nowait(data)
                elif type_bytes == META:
                    meta_data = json.loads(data)
                    #if meta_data.get("type") == "exit":
                    #    await self.send_meta({"type": "exit"})
                    #else:
                    #    await on_meta(json.loads(data))



from .button import Button as Button
from .plotext import Plotext as Plotext
from .switch import Switch as Switch
from .widget import Widget as Widget
from ._driver import Driver as Driver

__version__ = "0.5.4"



from __future__ import annotations

from ypywidgets import Reactive, Widget
from ypywidgets.comm import CommWidget


class PlotextModel(Widget):

    _clear_data = Reactive[bool](False)
    _args = Reactive[list]([])
    _scatter = Reactive[bool](False)
    _plot = Reactive[bool](False)
    _title = Reactive[str]("")

    def __init__(self, ydoc=None) -> None:
        super().__init__(ydoc)

    def clear_data(self) -> None:
        self._clear_data = False
        self._clear_data = True

    def scatter(self, *args) -> None:
        self._args = list(args)
        self._scatter = False
        self._scatter = True

    def plot(self, *args) -> None:
        self._args = list(args)
        self._plot = False
        self._plot = True

    def title(self, value: str) -> None:
        self._title = value


class Plotext(CommWidget, PlotextModel):

    def __init__(self) -> None:
        CommWidget.__init__(self)
        PlotextModel.__init__(self)



from textual.widgets import Button as TextualButton

from ..button import ButtonModel


class Button(TextualButton):

    def __init__(self, model: ButtonModel) -> None:
        super().__init__()
        self.ymodel = model

        @ButtonModel._press.watch
        def _watch__press(obj, old, new):
            if obj != model:
                return

            self.action_press()

        @ButtonModel.label.watch
        def _watch_label(obj, old: str, new: str):
            if obj != model:
                return

            self._set_label(new)

        @ButtonModel.variant.watch
        def _watch_variant(obj, old: str, new: str):
            if obj != model:
                return

            self._set_variant(new)

        @ButtonModel.disabled.watch
        def _watch_disabled(obj, old: bool, new: bool):
            if obj != model:
                return

            self._set_disabled(new)

        self._set_label(model.label)
        self._set_variant(model.variant)
        self._set_disabled(model.disabled)

    def on_button_pressed(self, event) -> None:
        self.ymodel._pressed = not self.ymodel._pressed

    def _set_label(self, value: str) -> None:
        if value != self.label:
            self.label = value

    def _set_variant(self, value: str) -> None:
        if value != self.variant:
            self.variant = value

    def _set_disabled(self, value: str) -> None:
        if value != self.disabled:
            self.disabled = value



import asyncio
from functools import lru_cache
from typing import Callable

import pyte
from rich.color import Color, parse_rgb_hex
from rich.console import RenderableType
from rich.segment import Segment
from rich.style import Style
from textual import events
from textual.geometry import Region
from textual.strip import Strip
from textual.widget import Widget as TextualWidget
from wcwidth import wcwidth as _wcwidth

from ..widget import WidgetModel


wcwidth: Callable[[str], int] = lru_cache(maxsize=4096)(_wcwidth)


CTRL_KEYS = {
    "left": "\u001b[D",
    "right": "\u001b[C",
    "up": "\u001b[A",
    "down": "\u001b[B",
}


class PyteDisplay:
    def __init__(self, screen):
        self._screen = screen
        self._cache = {}
        self._dirty = set()

    def get_line(self, y: int) -> Strip:
        self._dirty.update(self._screen.dirty)
        self._screen.dirty.clear()
        if y in self._dirty or y not in self._cache:
            line = self._screen.buffer[y]
            is_wide_char = False
            segments = []
            for x in range(self._screen.columns):
                if is_wide_char:
                    is_wide_char = False
                    continue
                char = line[x].data
                assert sum(map(wcwidth, char[1:])) == 0
                is_wide_char = wcwidth(char[0]) == 2
                char = line[x]
                color = None if char.fg == "default" else Color.from_triplet(parse_rgb_hex(char.fg))
                bgcolor = None if char.bg == "default" else Color.from_triplet(parse_rgb_hex(char.bg))
                segments.append(
                    Segment(
                        char.data,
                        Style(
                            color=color,
                            bgcolor=bgcolor,
                            bold=char.bold,
                            italic=char.italics,
                            underline=char.underscore,
                            blink=char.blink,
                            strike=char.strikethrough,
                            reverse=char.reverse,
                        )
                    )
                )
            self._cache[y] = Strip(segments)
            if y in self._dirty:
                self._dirty.remove(y)
        return self._cache[y]


class Terminal(TextualWidget, can_focus=True):
    DEFAULT_CSS = """
    Terminal {
        height: 1fr;
    }
    """

    def __init__(self, send_queue, recv_queue):
        super().__init__()
        self._send_queue = send_queue
        self._recv_queue = recv_queue
        self._screen = pyte.Screen(0, 0)
        self._stream = pyte.Stream(self._screen)
        self._display = PyteDisplay(self._screen)
        self._tasks = [asyncio.create_task(self._recv())]

    def render_line(self, y: int) -> Strip:
        return self._display.get_line(y)

    def on_resize(self, event: events.Resize):
        self.do_resize(event.size.width, event.size.height)

    def do_resize(self, cols: int, rows: int):
        if self.ymodel._cols != cols:
            self.ymodel._cols = cols
        if self.ymodel._rows != rows:
            self.ymodel._rows = rows
        self.ymodel._ready = False
        self.ymodel._ready = True
        self._screen = pyte.Screen(cols, rows)
        self._stream = pyte.Stream(self._screen)

    async def on_event(self, event: events.Event) -> None:
        await super().on_event(event)
        if isinstance(event, events.Key):
            char = CTRL_KEYS.get(event.key) or event.character
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return
        if isinstance(event, events.MouseMove):
            char = f"\x1b[<35;{event.x};{event.y}M"
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return
        if isinstance(event, events.MouseDown):
            char = f"\x1b[<0;{event.x};{event.y}M"
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return
        if isinstance(event, events.MouseUp):
            char = f"\x1b[<0;{event.x};{event.y}m"
            await self._send_queue.put("")
            await self._send_queue.put(char)
            event.stop()
            return

    async def _recv(self):
        while True:
            data = await self._recv_queue.get()
            if not data:
                continue
            self._stream.feed(data)
            self._display = PyteDisplay(self._screen)
            for y in self._screen.dirty:
                region = Region(0, y, self._screen.columns, 1)
                self.refresh(region)


class Widget(Terminal):

    def __init__(self, model: WidgetModel) -> None:
        self.ymodel = model
        self._data_from_app_queue = asyncio.Queue()
        self._data_to_app_queue = asyncio.Queue()
        super().__init__(self._data_to_app_queue, self._data_from_app_queue)

        @WidgetModel._data_from_app.watch
        def _watch__data_from_app(obj, old: str, new: str):
            if obj != model:
                return

            self._to_terminal(new)

        self._tasks = [asyncio.create_task(self._from_terminal())]

    def _to_terminal(self, data: str) -> None:
        self._data_from_app_queue.put_nowait(data)

    async def _from_terminal(self):
        while True:
            data = await self._data_to_app_queue.get()
            self.ymodel._data_to_app = data



from textual_plotext import PlotextPlot

from ..plotext import PlotextModel


class Plotext(PlotextPlot):

    def __init__(self, model: PlotextModel) -> None:
        super().__init__()
        self.ymodel = model

        @PlotextModel._clear_data.watch
        def _watch__clear_data(obj, old, new):
            if obj != model:
                return

            self.clear_data()

        @PlotextModel._scatter.watch
        def _watch__scatter(obj, old, new):
            if obj != model:
                return

            self.scatter()

        @PlotextModel._plot.watch
        def _watch__plot(obj, old, new):
            if obj != model:
                return

            self.plot()

        @PlotextModel._title.watch
        def _watch__title(obj, old: str, new: str):
            if obj != model:
                return

            self.title(new)

        self.clear_data()
        self.scatter()
        self.plot()
        self.title(model._title)

    def clear_data(self) -> None:
        if self.ymodel._clear_data:
            self.plt.clear_data()
            self.refresh()

    def title(self, value: str) -> None:
        self.plt.title(value)
        self.refresh()

    def scatter(self) -> None:
        if self.ymodel._scatter:
            self.plt.clear_data()
            self.plt.scatter(*self.ymodel._args)
            self.refresh()

    def plot(self) -> None:
        if self.ymodel._plot:
            self.plt.clear_data()
            self.plt.plot(*self.ymodel._args)
            self.refresh()



from __future__ import annotations

from typing import Callable

from ypywidgets import Reactive, Widget
from ypywidgets.comm import CommWidget


class ButtonModel(Widget):

    label = Reactive[str]("")
    variant = Reactive[str]("default")
    disabled = Reactive[bool](False)
    _press = Reactive[bool](False)
    _pressed = Reactive[bool](False)

    def __init__(
        self,
        label: str | None = None,
        variant: str | None = None,
        disabled: bool | None = None,
        on_button_pressed: Callable[[], None] | None = None,
        ydoc=None,
    ) -> None:
        super().__init__(ydoc)
        if label is not None:
            self.label = label
        if variant is not None:
            self.variant = variant
        if disabled is not None:
            self.disabled = disabled
        if on_button_pressed is None:
            def no_op():
                pass
            self._on_button_pressed = no_op
        else:
            self._on_button_pressed = on_button_pressed

    @property
    def on_button_pressed(self) -> Callable[[], None]:
        return self._on_button_pressed

    @on_button_pressed.setter
    def on_button_pressed(self, value: Callable[[], None]) -> None:
        self._on_button_pressed = value

    def action_press(self):
        self._press = not self._press

    @_pressed.watch
    def _watch__pressed(self, old: bool, new: bool):
        self._on_button_pressed()


class Button(CommWidget, ButtonModel):

    def __init__(
        self,
        label: str = "",
        variant: str = "default",
        disabled: bool = False,
        on_button_pressed: Callable[[], None] | None = None,
    ):
        CommWidget.__init__(self)
        ButtonModel.__init__(self, label, variant, disabled, on_button_pressed)
